"""
See "Tabu search for the cyclic bandwidth problem" by Rodriguez-Tello, Ramirez-Torres and Lardeux,
     Computers & Operations Research, 2015
See https://www.tamps.cinvestav.mx/~ertello/cbmp.php
"""

from pycsp3 import *

n, edges = data or (5, [[0, 1], (0, 4), (1, 2), (1, 3), (2, 3), (2, 4)])
edges = [tuple(t) for t in edges]  # because from JSON, we get lists and not tuples (which may be a problem with some conditions)

# x[i] is the label of the ith node
x = VarArray(size=n, dom=range(n))

satisfy(
    AllDifferent(x)
)

if not variant():
    minimize(
        Maximum(min(abs(x[i] - x[j]), n - abs(x[i] - x[j])) for i, j in edges)
    )


elif variant("aux"):

    # y[i][j] is the distance between the labels of nodes i and j (if an edge exists)
    y = VarArray(size=[n, n], dom=lambda i, j: range(1, n) if (i, j) in edges else None)

    satisfy(
        # computing distances
        y[i][j] == abs(x[i] - x[j]) for i, j in edges
    )

    minimize(
        Maximum(min(y[i][j], n - y[i][j]) for i, j in edges)
    )

"""
With an aggressive ub, optimality is proved:
  java ace CyclicBandwidth-path300.xml -ale=4 -ub=3
  java ace CyclicBandwidth-aux-path300.xml -ub=2
  
more efficient with -ale=4 on some instances (e.g. caterpillar)
"""
