"""
This problem is about optimizing the scheduling of filter operations, commonly used in High-Level Synthesis.
This model has been originally written by Krzysztof Kuchcinski for the 2010, 2012, 2013 and 2016 Minizinc competitions.

See also the models in JaCop -- https://github.com/radsz/jacop/tree/develop/src/main/java/org/jacop/examples/fd/filters
"""

from pycsp3 import *

del_add, del_mul, number_add, number_mul, last, add, dependencies = data
nOperations = len(dependencies)

d = [del_add if i in add else del_mul for i in range(nOperations)]
mul = [i for i in range(nOperations) if i not in add]

# t[i] is the starting time of the ith operation
t = VarArray(size=nOperations, dom=range(101))

# r[i] is the (index of the) operator used for the ith operation
r = VarArray(size=nOperations, dom=lambda i: range(1, 1 + (number_add if i in add else number_mul)))

satisfy(
    # respecting dependencies
    [t[i] + d[i] <= t[j] for i in range(nOperations) for j in dependencies[i]],

    # no overlap concerning add operations
    NoOverlap(origins=[(t[i], r[i]) for i in add], lengths=[(del_add, 1) for i in add]),

    # no overlap concerning mul operations
    NoOverlap(origins=[(t[i], r[i]) for i in mul], lengths=[(del_mul, 1) for i in mul])
)

minimize(
    # minimizing the ending time of last operations
    Maximum(t[i] + d[i] for i in last)
)
