"""
Minizinc 2017
"""

from pycsp3 import *


agents, horizon, neighbors, = data
src, dst = zip(*agents)
nAgents, nNodes = len(agents), len(neighbors)

assert all(agent[0] != agent[1] for agent in agents)

if variant("table"):

    table = [(i, v) for i, t in enumerate(neighbors) for v in t]

    # x[t][a] is the node where is the agent a at time t
    x = VarArray(size=[horizon + 1, nAgents], dom=range(nNodes))

    # e[a] is the time when the agent a arrives at its destination
    e = VarArray(size=nAgents, dom=range(horizon + 1))

    satisfy(
        # agents must occupy different node at any time
        [AllDifferent(x[t]) for t in range(horizon + 1)],

        # agents at their destinations stays there
        [(x[t][a] != dst[a]) | (x[t + 1][a] == dst[a]) for t in range(horizon) for a in range(nAgents)],

        # agents can only move to connected nodes
        [(x[t][a], x[t + 1][a]) in table for t in range(horizon) for a in range(nAgents)],

        # setting agents at their initial positions
        [x[0][a] == src[a] for a in range(nAgents)],

        # setting agents at their final positions
        [x[-1][a] == dst[a] for a in range(nAgents)],

        # computing end times of agents
        [(e[a] == t + 1) == ((x[t][a] != dst[a]) & (x[t + 1][a] == dst[a])) for t in range(horizon) for a in range(nAgents)]
    )

    if subvariant("mks"):
        minimize(
            Maximum(e)
        )
    else:
        minimize(
            Sum(e) + nAgents
        )

"""
1) using SAC3 as preprocessing allows us to solve efficiently the 5 minizinc instances for variant 'table'

2) [(B[t, a, v] == 0) | (Sum(B[t + 1, a, u] for u in rel[v]) >= 1) for t in range(makespan) for a in range(k) for v in range(n)]

"""
