"""
See, e.g., "Constraint programming lessons learned from crossword puzzles" by A. Beacham, X. Chen, J. Sillito, and P. van Beek. 2001

Examples of Execution:
  python3 Crossword.py -data=Crossword_vg6-7-ogd.json
"""

from pycsp3 import *

spots, dict_name = data
words = dict()
for line in open(dict_name):
    code = alphabet_positions(line.strip().lower())
    words.setdefault(len(code), []).append(code)


def find_holes(matrix, transposed):
    def build_hole(row, col, size, horizontal):
        return Hole(row, slice(col, col + size), size) if horizontal else Hole(slice(col, col + size), row, size)

    Hole = namedtuple("Hole", "i j r")  # i and j are indexes (one of them being a slice) and r is the size
    p, q = len(matrix), len(matrix[0])
    t = []
    for i in range(p):
        start = -1
        for j in range(q):
            if matrix[i][j] == 1:
                if start != -1 and j - start >= 2:
                    t.append(build_hole(i, start, j - start, not transposed))
                start = -1
            elif start == -1:
                start = j
            elif j == q - 1 and q - start >= 2:
                t.append(build_hole(i, start, q - start, not transposed))
    return t


holes = find_holes(spots, False) + find_holes(columns(spots), True)
n, m, nHoles = len(spots), len(spots[0]), len(holes)


#  x[i][j] is the letter, number from 0 to 25, at row i and column j (when no spot)
x = VarArray(size=[n, m], dom=lambda i, j: range(26) if spots[i][j] == 0 else None)

satisfy(
    # fill the grid with words
    x[i, j] in words[r] for (i, j, r) in holes   
)


""" Comments
1) we use lists instead of sets for tables ([(i1, i2)  ... instead of  {(i1, i2) ..) because it is quite faster to process

2) it is not possible to write x[i][j] when i is a slice; this must be x[i, j]
"""
